/* 
 * File:     cAlert.h
 * Project:  LinWarrior 3D
 * Home:     hackcraft.de
 *
 * Created on February 14, 2010, 7:02 PM
 */

#ifndef _CALERT_H
#define	_CALERT_H

#include "cObject.h"
#include "cParticle.h"

#include <string>
#include <hash_set>
#include <set>

/**
 * Checks a defined area for certain intruding objects
 * (most likely the player). When such an intrusion is
 * detected a message is sent to a given group of objects.
 * Those receivers may then react on that message.
 *
 * Some ideas:
 * - Enemies powering up as soon as you enter a zone (pop WAIT state).
 * - Enemies collectively attacking the intruder (push ATTACK target).
 * - Allied backup moving in after you (push GOTO destination).
 * - Joining you when reaching a roundevous point (push FOLLOW target).
 * - Warning and mission abort when leaving mission area.
 *
 * Some further ideas:
 * Message aggregation and processing objects for and, or, xor, forwarding,
 * would allow extended mission scripting.
 * Likewise switching alerts on and off would present opportunities.
 *
 */
class cAlert : public cObject {
public:
    struct rShape {

        enum Shapes {
            CYLINDER, BOX, SPHERE, MAX_SHAPES
        };
        int type;
        float center[3];
        float range[3];
    };
private:
    rShape shape;
    std::string message;
    OID group;
    std::set<OID> intruders;
    /// Use positive shape (true) or negative of shape (false) as zone?
    bool positive;
    /// Fire message when entering (true) or exiting (false) the zone?
    bool posedge;
    /// Fire message only once (true) or always when conditions are true?
    bool once;
    /// Set true when the message was fired at least once - initially false.
    bool fired;
    /// Reacting only on objects with such roles:
    std::set<OID> sensitivity;
public:

    cAlert(float* center, float* range, int shapetype, std::string message, OID group, bool positive = true, bool posedge = true, bool once = false) {
        vector_cpy(shape.center, center);
        vector_cpy(shape.range, range);
        this->traceable->radius = 1 + fmax(fabs(shape.range[0]), fmax(fabs(shape.range[1]), fabs(shape.range[2])));
        if (this->traceable->radius >= 25) {
            // Make global "position".
            vector_set(traceable->pos, float_NAN, float_NAN, float_NAN);
        } else {
            vector_cpy(traceable->pos, center);
        }
        this->positive = positive;
        this->posedge = posedge;
        this->once = once;
        this->fired = false;
        this->message = message;
        this->group = group;
        collideable = new rCollideable;
        damageable = new rDamageable;
        addRole(COLLIDEABLE);
        addRole(DAMAGEABLE);
        shape.type = shapetype;
        sensitivity.insert(HUMANPLAYER);
    }

    virtual float constrainParticle(float* worldpos, float radius = 0.0f, float* localpos = NULL, cObject* enactor = NULL) {
        //std::cout << "called" << std::endl;
        // Don't react on scanning and shooting.
        if (enactor == NULL) return 0;
        if (once && fired) return 0;
        if (! enactor->allRoles(&sensitivity) ) return 0;
        
        rShape* s = &shape;
        float depth = 0;
        float projection[3];
        switch (s->type) {
            case rShape::CYLINDER:
            {
                depth = cParticle::constraintParticleByCylinder(worldpos, shape.center, shape.range[2], shape.range[1], projection);
                break;
            }
            case rShape::BOX:
            {
                float min[3];
                float max[3];
                vector_set3i(min, shape.center[i] - shape.range[i]);
                vector_set3i(max, shape.center[i] + shape.range[i]);
                depth = cParticle::constraintParticleByAABB(worldpos, min, max, projection);
                break;
            }
            case rShape::SPHERE:
            {
                depth = cParticle::constraintParticleBySphere(worldpos, shape.center, shape.range[0], projection);
                break;
            }
            default:
                break;
        }
        //std::cout << depth << std::endl;
        // Is that possible intruder yet unknown?
        if (intruders.find(enactor->base->oid) == intruders.end()) {
            // The intruder was outside and now there is an intrusion?
            if ((depth != 0) == positive) {
                intruders.insert(enactor->base->oid);
                if (posedge) {
                    fired = true;
                    cWorld::instance->sendMessageT(base->oid, group, message, NULL);
                }
            }
        } else  {
            // The intruder was inside and now the intruder left the zone?
            if ((depth == 0) == positive) {
                intruders.erase(enactor->base->oid);
                if (!posedge) {
                    fired = true;
                    cWorld::instance->sendMessageT(base->oid, group, message, NULL);
                }
            }
        }
        return 0;
    }
};

#endif	/* _CALERT_H */

