#include "cController.h"

#include "cObject.h"
#include "cWorld.h"

#include <cstdlib>
#include <cassert>
#include <iostream>

using std::cout;
using std::endl;
using std::string;

#define debug_state !true
#define debug_transitions !true
#define debug_step !true

// -------------------------------------------------------------

cController::cController(cObject* entity, bool enabled) {
    mDevice = entity;
    mEnabled = enabled;
}

cController::~cController() {
    while (!mStack.empty()) mStack.pop_back();
}

void cController::printState() {
    int framesize = getFrameSize();
    string framename = getFrameName();
    cout << "CtrlSt of " << mDevice->nameable->name.c_str() << "#" << mDevice->base->oid << "  Frame: " << framename << "  Framesize:" << framesize << "  Stack: " << mStack.size() << endl;
}

string cController::getFrameName() {
    //cout << "getFrameName()\n";
    const char* names[] = {
        "WAIT", "ATTACK", "FOLLOW", "GOTO", "REPEAT",
        "NOSTATENAME"
    };
    unsigned int i = mStack.back();
    i = (i < OPCODE_MAX) ? i : OPCODE_MAX - 1;
    string s = string(names[i]);
    return s;
};

unsigned int cController::getFrameSizeOf(int opcode) {
    switch (opcode) {
        case WAIT: return 3;
        case ATTACK: return 2;
        case FOLLOW: return 3;
        case GOTO: return 5;
        case REPEAT : return 2;
        default: return 2;
    }
}

unsigned int cController::getFrameSize() {
    return getFrameSizeOf(mStack.back());
}

OID cController::getParameter(int offset) {
    return mStack[mStack.size() - offset - 1];
}

void cController::setParameter(int offset, OID value) {
    mStack[mStack.size() - offset - 1] = value;
}

void cController::push(OID value) {
    mStack.push_back(value);
}

void cController::pop() {
    if (debug_transitions) {
        cout << mDevice->nameable->name.c_str() << "#" << mDevice->base->oid << ".pop()\n";
    }
    int size = getFrameSize(); // Important: eval outside loop-condition!
    loopi(size) mStack.pop_back();
    if (debug_transitions) {
        printState();
    }
}

void cController::process(float spf) {
    if (debug_state) cout << "cController::process()\n";

    if (!mEnabled) return;
    if (mDevice == NULL) return;

    if (mStack.empty()) pushWaitEvent();

    if (debug_step) printState();

    switch (mStack.back()) {
        case WAIT: waitEvent();
            break;
        case ATTACK: attackEnemy();
            break;
        case FOLLOW: followLeader();
            break;
        case GOTO: gotoDestination();
            break;
        case REPEAT: repeatInstructions();
            break;
        default: cout << "Invalid Instruction Request!\n";
            break;
    }
    if (debug_step) {
        cout << "=> ";
        printState();
    }
}


// -------------------------------------------------------------

void cController::pushWaitEvent(long mseconds, bool patrol) {
    push(patrol);
    push(mseconds);
    push(WAIT);
}

void cController::waitEvent() {
    //OID opcode = getParameter(0);
    long mseconds = getParameter(1);
    bool patrol = getParameter(2);
    
    {
        mDevice->do_idle();
    }

    if (patrol) {
        OID enemy = 0;
        if (enemy == 0) enemy = mDevice->disturbedBy();
        if (enemy == 0) enemy = mDevice->enemyNearby();
        if (enemy) {
            {
                OID self = mDevice->base->oid;
                cWorld::instance->sendMessage(0, 0, "#%llu: Intruder!\n", self);
            }
            pushAttackEnemy(enemy);
        }
    }

    if (mseconds > 0) {
        mseconds -= 1000/40;
        setParameter(1, mseconds);
        if (mseconds <= 0) {
            pop();
            return;
        }
    }
}

// -------------------------------------------------------------

void cController::pushAttackEnemy(OID entity) {
    if (debug_transitions) {
        cout << mDevice->nameable->name.c_str() << "#" << mDevice->base->oid << ".pushAttackEnemy( " << entity << " )\n";
    }
    {
        //OID self = mDevice->mSerial;
        //cWorld::instance->sendMessage("@%llu: All ur base belongs to us.\n", entity);
    }
    push(entity);
    push(ATTACK);
}
 
void cController::attackEnemy() {
    //OID opcode = getParameter(0);
    OID entity = getParameter(1);

    {
        mDevice->do_moveFor(NULL);
        mDevice->do_aimFor(entity);
        mDevice->do_moveNear();
        mDevice->do_aimAt();
        if (mDevice->inWeaponRange() > 0.5) mDevice->do_fireAt();
        //((cMech*)mDevice)->Pattern(tf, "nrnlln");
    }
    
    cObject* target = cWorld::instance->mIndex[entity];
    if (mDevice->inTargetRange() < 0.01) {
        mDevice->do_aimFor(NULL);
        pop();
        return;
    }
    if (target == NULL) { // Target disappeared
        mDevice->do_aimFor(NULL);
        pop();
        return;
    } else if (target->hasRole(DEAD)) {
        mDevice->do_aimFor(NULL);
        pop();
        return;
    }

}

// -------------------------------------------------------------

void cController::pushFollowLeader(OID entity, bool patrol) {
    if (debug_transitions) {
        cout << mDevice->nameable->name.c_str() << "#" << mDevice->base->oid << ".pushFollowLeader( " << entity << ", " << patrol << " )\n";
    }
    push(patrol ? 1 : 0);
    push(entity);
    push(FOLLOW);
}

void cController::followLeader() {
    //OID opcode = getParameter(0);
    OID entity = getParameter(1);
    OID patrol = getParameter(2);

    {
        mDevice->do_aimFor(entity);
        mDevice->do_moveFor(NULL);
        mDevice->do_moveNear();
        mDevice->do_aimAt();
    }

    if (patrol) {
        OID nearby = mDevice->enemyNearby();
        if (nearby) {
            pushAttackEnemy(nearby);
            return;
        }
    }
}

// -------------------------------------------------------------

void cController::pushGotoDestination(float* v, bool patrol) {
    if (v == NULL) return;
    if (debug_transitions) {
        cout << mDevice->nameable->name.c_str() << "#" << mDevice->base->oid << ".pushGotoDestination( <" << v[0] << "," << v[1] << "," << v[2] << ">, " << patrol << " )\n";
    }
    unsigned long *p = (unsigned long*) v;
    push(patrol ? !0 : 0);
    push(p[2]);
    push(p[1]);
    push(p[0]);
    push(GOTO);
}

void cController::gotoDestination() {
    //OID opcode = getParameter(0);
    OID v0 = getParameter(1);
    OID v1 = getParameter(2);
    OID v2 = getParameter(3);
    OID patrol = getParameter(4);
    unsigned long p[] = {
        v0, v1, v2
    };
    float* v = (float*) p;

    {
        if (debug_state) cout << "going " << ((patrol > 0) ? "patrolling" : "directly") << " to <" << v[0] << "," << v[1] << "," << v[2] << " >\n";
        mDevice->do_moveFor(v);
        //mDevice->do_aimFor(0);
        mDevice->do_aimAt();
        mDevice->do_moveTowards();
    }

    float range = mDevice->inDestinationRange();
    if (debug_state) cout << "DestinationRange " << range << endl;
    if (range > 0.0f) {
        pop();
        return;
    }

    if (patrol) {
        OID nearby = mDevice->enemyNearby();
        if (nearby) {
            pushAttackEnemy(nearby);
            return;
        }
    }
}

// -------------------------------------------------------------

void cController::pushRepeatInstructions(int n) {
    if (debug_transitions) {
        cout << mDevice->nameable->name.c_str() << "#" << mDevice->base->oid << ".pushRepeatInstructions( " << n << " )\n";
    }
    push(n);
    push(REPEAT);
}

void cController::repeatInstructions() {
    OID opcode = getParameter(0);
    OID n = getParameter(1);

    if (debug_state) {
        printState();
    }

    unsigned int first = getFrameSize();
    if (debug_state) cout << "first " << first << " -> opcode " << opcode << endl;
    unsigned int last = first;
    loopi(n) {
        int opcode = getParameter(last);
        last += getFrameSizeOf(opcode);
        if (debug_state) cout << "last " << last << " -> opcode " << opcode << endl;
    }

    int size = last - first;
    if (debug_state) cout << size << " = " << last << " - " << first << endl;

    loopi(size) {
        push(getParameter(last - 1));
    }
}


