/**
 * File:     cMech.h
 * Project:  LinWarrior 3D
 * Home:     hackcraft.de
 *
 * Created on March 28, 2008, 10:23 PM (Dez1999)
 */

#ifndef _CMECH_H
#define	_CMECH_H

class cMech;

#include "psi3d/md5frmt.h"

#include "cObject.h"
#include "cWeapon.h"
#include "cComcom.h"

#include <vector>
#include <map>

class cWeapon;

// Gamepad Mapping
#define MECH_CHASSIS_LR_AXIS  AX_LR1
#define MECH_TURRET_LR_AXIS   AX_LR2
#define MECH_TURRET_UD_AXIS   AX_UD2
#define MECH_THROTTLE_AXIS    AX_UD1
//#define MECH_FIRE_BUTTON      BT_CROSS
#define MECH_FIRE_BUTTON      BT_R2
//#define MECH_JET_BUTTON       BT_SQUARE
#define MECH_JET_BUTTON       BT_L2
#define MECH_CAMERA_BUTTON    BT_DELTA

/**
 * Models Mechlike Objects.
 * 
 */
class cMech : public cObject {
    friend class cWepcom;
    friend class cForcom;
protected:


    /// # of camera modes: ISO C++ forbids const int with initialization >:(

    enum {
        MAX_CAMERAMODES = 6
    };


protected:

    /// Instance counter.
    static int sInstances;

    /// Texture Binds shared between instances.
    static std::map<int,long> sTextures;

protected:
    std::vector<cWeapon*> mWeapons;

    int mCurrentWeapon;

    cExplosion mExplosion;

    // Base angles in radians.
    float mBse[3];

    // Tower angles in radians.
    float mTwr[3];

    // Current jumpjet set-point.
    float mJetOn;

    // Current throttle set-point.
    float mThrottle;

    /// Current Camera mode, negative number is indicating transition.
    int mCameraState;

    struct rComputerised {
        cComcom* comcom;
        cTarcom* tarcom;
        cSyscom* syscom;
        cWepcom* wepcom;
        cForcom* forcom;
        cNavcom* navcom;
        /// Constructor
        rComputerised(cObject * o);
        /// Destructor
        ~rComputerised();
        /// Time step computers.
        void process(float dt);
    };

    rComputerised* computerised;

    struct rRigged {
        /// Model scale.
        float scale;
        /// The "static" model just as it is loaded.
        MD5Format::model* model;
        /// Actual local-(model-)space joints for this instance.
        MD5Format::joint* joints;
        /// Maps jointpoint identifier to actual joint index of the model.
        std::map<int, int> jointpoints;
        /// Joint angles for animation.
        std::map<int, std::map<int, float> > rotators;
        // Mech Joint and Mountpoint Matrices.
        float mHead[16];
        float mCTMount[16];
        float mBKMount[16];
        float mRSMount[16];
        float mRAMount[16];
        float mLSMount[16];
        float mLAMount[16];
        //
        std::map<int, float*> baseverts;
        /// Enumeration for indexing joints in animation.

        enum Jointpoints {
            EYE, HEAD, NECK,
            CTMOUNT, LAMOUNT, RAMOUNT, LSMOUNT, RSMOUNT, BKMOUNT,
            JET0, JET1, JET2, JET3, JET4,
            SPINE, TORSOR, LEFTLEG, RIGHTLEG, LEFTCALF, RIGHTCALF, MAX_JOINTPOINTS
        };

        /// Constructor

        rRigged() : scale(1.0f), model(NULL), joints(NULL) {
        }
        /// Destructor

        ~rRigged() {
            delete model;
            delete joints;
        }

        std::string getJointname(unsigned int num) {
            const char* names[] = {
                "EYE", "HEAD", "NECK",
                "CTMOUNT", "LAMOUNT", "RAMOUNT", "LSMOUNT", "RSMOUNT", "BKMOUNT",
                "JET0", "JET1", "JET2", "JET3", "JET4",
                "SPINE", "TORSOR", "LEFTLEG", "RIGHTLEG", "LEFTCALF", "RIGHTCALF"
            };
            if (num >= MAX_JOINTPOINTS) return string("");
            return string(names[num]);
        }
    };

    rRigged* rigged;


public:
    cMech(float* pos = NULL, float* rot = NULL);
    ~cMech();

    // Events
    virtual void onMessage(void* message);
    virtual void onSpawn();

    // Camera and Headphone
    void multEyeMatrix();
    void setAsAudioListener();

    // Physics and posing accordingly
    void animatePhysics(float spf);
    void poseRunning(float spf);
    void poseJumping(float spf);

    // Steering
    void ChassisLR(float radians);
    void ChassisUD(float radians);
    float TowerLR(float radians);
    void TowerUD(float radians);

    // Weapons
    void fireAllWeapons();
    void fireWeapon(unsigned n);
    void mountWeapon(char* point, cWeapon *weapon, bool add = true);

    // World Step
    virtual void animate(float spf);
    virtual void transform();
    virtual void drawSolid();
    virtual void drawEffect();
    virtual void drawHUD();

    // Particle constraining
    virtual void damageByParticle(float* localpos, float damage, cObject* enactor = NULL);
    virtual float constrainParticle(float* worldpos, float radius = 0.0f, float* localpos = NULL, cObject* enactor = NULL);
    float constrainParticleToWorld(float* worldpos, float radius = 0.0f);
public:
    // Queries
    virtual OID enemyNearby();
    virtual OID disturbedBy();

    // Fuzzy Predicates [0,1]
    virtual float inDestinationRange();
    virtual float inMeeleRange();
    virtual float inWeaponRange();
    virtual float inTargetRange();

    // Output Actions
    virtual void do_moveTowards();
    virtual void do_moveNear();
    virtual void do_aimAt();
    virtual void do_fireAt();
    virtual void do_idle();
    virtual void do_aimFor(OID target);
    virtual void do_moveFor(float* dest);
};


#endif	/* _CMECH_H */

