/**
 * File:     cWeapon.h
 * Project:  LinWarrior 3D
 * Home:     hackcraft.de
 *
 * Created on March 5, 2008, 4:09 PM
 */

#ifndef _CWEAPON_H
#define	_CWEAPON_H

class cWeapon;
#include "cWorld.h"
#include <list>
#include <vector>

#include "cParticle.h"

/**
 * Base-Class of all weapons.
 * Defines "all you need" for designing different kinds of weapons and
 * their visual effects. Note that weapons are themselves responsible for
 * drawing themselves, their bullets, missiles, dust, smoke and other particles.
 */
class cWeapon {
public:
    /// Object whielding the weapon (normaly not damaged by weapon).
    cObject* mOwner;

    /// Points To Current Vehicle Base Matrix of owner.
    float* mBasefv;

    /// Mounting Base Matrix relative to Vehicle Base Matrix or owner.
    float* mMountfv;

    /// Current Matrix indicating world pos & orientation (pose) of this weapon.
    float mPosef[16];

    /// Relative size and power of the weapon, 1.0 default mechscale weapon.
    float mScale;

    /// Remaining reload time until next firing is possible.
    float mReloading;

    /// Remaining readying time for bringing up the weapon (unused).
    float mReadying;

    /// Remaining firing time when firing takes more than an instant time.
    float mFiring;

    /// Remaining shots until reloading another clip (negative == infinitive?).
    short mAmmo;

    /// Remaining clips (negative == infinitive?).
    short mClips;

    /// Maximum amount of Ammo inside a clip.
    short mClipsize;

    /// Maximum amount of clips.
    short mDepotsize;

    /// OpenAL sound source for this weapon's sounds.
    unsigned mSoundSource;

    std::list<cParticle*> mShrapnels;
    std::list<cParticle*> mCastoff;
    std::list<cParticle*> mMissiles;
    std::list<cParticle*> mDamages;

public:
    cWeapon();

    void playSource();
    void playSourceIfNotPlaying();

    /** "Eignungsbeurteilung"
     *
     * Fuzzy-Value [0,1] indicating usability or fittness
     * of this weapon to be fired right now (not considering weapon state).
     * May sum up (depends on Weapon) (future):
     * +Ability/Possibility to hit target (may roll some dice).
     * +Possible damage on target.
     * +Possible damage on other enemy units (owner's enemies).
     * -Possible self-damage.
     * -Possible ally-damage.
     * +In (optimal) Firerange.
     * -Out of (optimal) Firerange.
     */
    virtual float fit(OID target) {
        return 0.5f;
    };

    /**
     * Considering weapon state, true when weapon is
     * at hand, ready-loaded and unlocked.
     */
    virtual bool ready() {
        return !((mReloading > +0.0f)
                || (mReadying > +0.0f)
                || (mFiring > +0.0f)
                || (mAmmo == 0)
                );
    };

    /**
     * Fire Weapon at target, target may be 0.
     */
    virtual void fire(OID target) = 0;

    /**
     * Update weapon animation cycles, particle trajectories and timers.
     */
    virtual void animate(float spf) {
    };

    /**
     * Calculate transformations for painting called between animate and draw*.
     * Calculates current pose according to owners Base and Mount matrices
     * and sets position of sound-source.
     */
    virtual void transform();

    /**
     * Draw the weapon model itself and non-translucent particles.
     */
    virtual void drawSolid() {
    };

    /**
     * Draw translucent effects like weapon flares and particles.
     */
    virtual void drawEffect() {
    };

    /**
     * Paints an iconic display of the weapon type and
     * state into ortho [0..1,0..1] view.
     */
    virtual void drawHUD() {
    };

    /**
     * Returns amount of damaged objects (currently just one at most).
     */
    int damageByParticle(float* worldpos, float radius, int roles, float damage);
};

/**
 * Your usual Machine or Mini-Gun.
 */
class cMachineGun : public cWeapon {
public:
    cMachineGun();

    virtual void fire(OID target);
    virtual void animate(float spf);
    virtual void drawSolid();
    virtual void drawEffect();
    virtual void drawHUD();
};

/**
 * Energetic Particle Rail-Gun.
 */
class cPlasmaGun : public cWeapon {
public:
    cPlasmaGun();

    virtual void fire(OID target);
    virtual void animate(float spf);
    virtual void drawSolid();
    virtual void drawEffect();
    virtual void drawHUD();
};

/**
 * Big Flaming slow moving bullet type gun.
 */
class cSparkGun : public cWeapon {
public:
    cSparkGun();

    virtual void fire(OID target);
    virtual void animate(float spf);
    virtual void drawSolid();
    virtual void drawEffect();
    virtual void drawHUD();
};

/**
 * Laser Beam kind of Weapon.
 */
class cRaybeam : public cWeapon {
public:
    cRaybeam();

    virtual void fire(OID target);
    virtual void animate(float spf);
    virtual void drawSolid();
    virtual void drawEffect();
    virtual void drawHUD();
};

/**
 * Small swarming homing missile.
 */
class cHomingMissile : public cWeapon {
public:
    cHomingMissile();

    virtual void fire(OID target);
    virtual void animate(float spf);
    virtual void drawSolid();
    virtual void drawEffect();
    virtual void drawHUD();
};

/**
 * Explosion with dust, smoke, dirt and shrapnels.
 */
class cExplosion : public cWeapon {
public:
    cExplosion();

    virtual void fire(OID target);
    virtual void animate(float spf);
    virtual void drawSolid();
    virtual void drawEffect();
};


#endif	/* _CWEAPON_H */

