/* 
 * File:   noise.h
 * Author: Benjamin Pickhardt
 * Home:   hackcraft.de
 *
 * Created on March 9, 2010, 11:14 AM
 */

#ifndef _NOISE_H
#define	_NOISE_H

/*
 * Encapsulates algorithms for noise generation and general
 * procedural content generation.
 *
 * Noise is the Message.
 */

/**
 * Calculates a simplex noise value for a 3d voxel or a 2d pixel and 1d time.
 * To produce Cloud/Terrain like textures add several octaves.
 * By holding one or two dimension fixed you can extract lower dimensional
 * noise.
 *
 * Ref: Simplex-Noise-Demystified (read this)
 * Ref: Simplex Noise by Ken Perlin (not read)
 * Ref: "Improving Noise" by Ken Perlin (16 instead of 12 gradients)
 * I just implemented it adding permutation through LFSR and
 * with calculation of an edge-direction lookup index for a table instead.
 *
 * @param seed an octave unique seed value in [0,255].
 * @return float value in about [-1,+1].
 */
static float simplex3(float x, float y, float z, unsigned char seed = 0) {
    // Gradient Scaling (to fit end result in [-1,+1]).
    const float s = 32;
    // Gradients (first row with redundancy to make it 16).
    static float g[16][3] = {
        {+s,+s, 0}, {-s,+s, 0}, { 0,-s,+s}, { 0,-s,-s}, // redundant entries.
        {+s,+s, 0}, {-s,+s, 0}, {+s,-s, 0}, {-s,-s, 0},
        {+s, 0,+s}, {-s, 0,+s}, {+s, 0,-s}, {-s, 0,-s},
        { 0,+s,+s}, { 0,-s,+s}, { 0,+s,-s}, { 0,-s,-s}
    };
    // 3 times the permutation of values 0 to 255.
    // We use three times instead of two times so we can shift the start.
    // There are originally two times so that perm[x+perm[y]] does not
    // exceed the array bounds and no modulo (or &-ing) neccessary.
    // Permutation through 8-bit LFSR (with full 256 period: 0-255).
    static unsigned char perms[768];
    static bool init = true;
    if (init) {
        init = false;
        unsigned char b = 131;

        loopi(768) {
            unsigned char x8 = b >> 7;
            unsigned char x6 = b >> 5;
            unsigned char x5 = b >> 4;
            unsigned char x4 = b >> 3;
            unsigned char prod = (x8 ^ x6 ^ x5 ^ x4 ^ 1) & 1;
            b = (b << 1) | prod;
            perms[i] = b;
        }
    }
    // Select permutation start point through seed.
    unsigned char* p = &perms[seed];

    const float skew = 1.0f / 3.0f;
    const float unsk = 1.0f / 6.0f;

    // Transform orthogonal input to skewed simplex space.
    // Take the floor to get the simplex grid point.
    float offset = (x + y + z) * skew;
    int x_ = ({float f = (x + offset); (f > 0) ? (int) f : (int) (f - 1);});
    int y_ = ({float f = (y + offset); (f > 0) ? (int) f : (int) (f - 1);});
    int z_ = ({float f = (z + offset); (f > 0) ? (int) f : (int) (f - 1);});

    // And transform grid point back to orthogonal space.
    offset = (x_ + y_ + z_) * unsk;
    float x__ = x_ - offset;
    float y__ = y_ - offset;
    float z__ = z_ - offset;

    // Determine relative distance to grid point in orthogonal space.
    float xofs0 = x - x__;
    float yofs0 = y - y__;
    float zofs0 = z - z__;

    // Relative distance relations tell us in which of the 6 tetrahedron we are.
    // Then we need to walk along edges to find the corners.
    // For each of the six (=rows) there are 2 edge direction vectors.
    // These direction vectors are called "left" and "right from here on
    // also they are not neccessaryily the left and right edges.
    int edgedirections[][3] = {
        { 1, 0, 0}, {1, 1, 0}, // x >= y >= z
        { 1, 0, 0}, {1, 0, 1}, // x >= z >= y
        { 0, 0, 0}, {0, 0, 0}, // - (transitivity disallows this)
        { 0, 0, 1}, {1, 0, 1}, // z >= x >= y

        { 0, 1, 0}, {1, 1, 0}, // y >= x >= z
        { 0, 0, 0}, {0, 0, 0}, // - (transitivity disallows this)
        { 0, 1, 0}, {0, 1, 1}, // y >= z >= x
        { 0, 0, 1}, {0, 1, 1}, // z >= y >= x
    };

    // Calculate a tetrahedron index (eight hypothetical, just six meaningful).
    // The bits are shifted one left for 2 vectors on each row (see above).
    int order = 0;
    order |= (xofs0 >= yofs0) ? 0 : 8;
    order |= (xofs0 >= zofs0) ? 0 : 4;
    order |= (yofs0 >= zofs0) ? 0 : 2;

    int* left = edgedirections[order + 0];
    int* right = edgedirections[order + 1];

    // Simplex corner 2 offset ("left")
    float xofs1 = xofs0 - left[0] + 1.0f * unsk;
    float yofs1 = yofs0 - left[1] + 1.0f * unsk;
    float zofs1 = zofs0 - left[2] + 1.0f * unsk;
    // Simplex corner 3 offset ("right")
    float xofs2 = xofs0 - right[0] + 2.0f * unsk;
    float yofs2 = yofs0 - right[1] + 2.0f * unsk;
    float zofs2 = zofs0 - right[2] + 2.0f * unsk;
    // Simplex corner 4 offset ("far")
    float xofs3 = xofs0 - 1 + 3.0f * unsk;
    float yofs3 = yofs0 - 1 + 3.0f * unsk;
    float zofs3 = zofs0 - 1 + 3.0f * unsk;

    // Have the base grid point wrapped inside [0,255].
    int u = x_ & 255;
    int v = y_ & 255;
    int w = z_ & 255;

    // Get gradient index for this permutation - one out of 16 possible.
    // "Base" corner.
    int g0 = p[u + p[v + p[w]]] & 15;
    // "Left" corner.
    int g1 = p[u + left[0] + p[v + left[1] + p[w + left[2]]]] & 15;
    // "Right" corner.
    int g2 = p[u + right[0] + p[v + right[1] + p[w + right[2]]]] & 15;
    // "Far" corner.
    int g3 = p[u + 1 + p[v + 1 + p[w + 1]]] & 15;

    // Contribution "attenuation" factors for distances to grid corner points.
    float c0 = 0.6f - (xofs0*xofs0 + yofs0*yofs0 + zofs0*zofs0);
    float c1 = 0.6f - (xofs1*xofs1 + yofs1*yofs1 + zofs1*zofs1);
    float c2 = 0.6f - (xofs2*xofs2 + yofs2*yofs2 + zofs2*zofs2);
    float c3 = 0.6f - (xofs3*xofs3 + yofs3*yofs3 + zofs3*zofs3);

    // Contribution sum of the four corners.
    float sum = 0.0f;

    // base
    if (c0 > 0) {
        float* g_ = g[g0];
        float sqr = c0 * c0;
        sum += sqr * sqr * (g_[0]*xofs0 + g_[1]*yofs0 + g_[2]*zofs0);
    }

    // left
    if (c1 > 0) {
        float* g_ = g[g1];
        float sqr = c1 * c1;
        sum += sqr * sqr * (g_[0]*xofs1 + g_[1]*yofs1 + g_[2]*zofs1);
    }

    // right
    if (c2 > 0) {
        float* g_ = g[g2];
        float sqr = c2 * c2;
        sum += sqr * sqr * (g_[0]*xofs2 + g_[1]*yofs2 + g_[2]*zofs2);
    }

    // far
    if (c3 > 0) {
        float* g_ = g[g3];
        float sqr = c3 * c3;
        sum += sqr * sqr * (g_[0]*xofs3 + g_[1]*yofs3 + g_[2]*zofs3);
    }

    // [-1,+1]
    return sum;
}


#endif	/* _NOISE_H */

